from __future__ import annotations
from typing import Any
from excelbird.core.gap import Gap
from dataclasses import dataclass

@dataclass(slots=True)
class Locable:
    elem: ListIndexableById

    def __getitem__(self, key: slice) -> Any:
        if not isinstance(key, slice):
            return self.elem.__getitem__(key)

        start, stop = key.start, key.stop
        if start is None:
            start = 0
        if stop is None:
            stop = -1

        elem_from = self.elem[self.elem._key_to_idx(start)]
        elem_to = self.elem[self.elem._key_to_idx(stop)]
        return elem_from >> elem_to


class ListIndexableById(list):
    """
    A simple child class of list that can accept an `id` string as a
    key to access elements.

    Each element MUST have an `id` property itself, before trying to
    access elements.
    """

    @property
    def loc(self) -> Locable:
        return Locable(self)

    def insert(self, index, new) -> None:
        index = self._key_to_idx(index)
        super().insert(index, new)

    def set(self, **kwargs) -> ListIndexableById:
        for k, v in kwargs.items():
            setattr(self, k, v)
        return self

    def get(self, key, default=None) -> Any:
        """
        Safely get an element.

        Parameters
        ----------
        key : str or int
            The index, ``id`` or ``header`` (if series) of a child element.
        default : Any, default None
            Value to return if nothing is found

        Returns
        -------
        Any
            Note that some dynamic elements, such as :class:`Gap` or :class:`Expr` may not
            have been resolved to a valid child type yet.

        Notes
        -----

        .. note::

            Excelbird containers are all subclasses of :class:`python:list` so you
            can access elements using square brackets the same as you would with a list.

        """
        try:
            return self[key]
        except Exception:
            return default

    def _key_to_idx(self, key) -> int:
        if isinstance(key, int):
            return key

        ids = [i.id if hasattr(i, "_id") else None for i in self]
        if key in ids:
            return ids.index(key)
        else:
            headers = [
                i.header if hasattr(i, "_header") else 
                i.kwargs.get('header', None) if hasattr(i, 'kwargs')
                else None for i in self
            ]
            if key in headers:
                return headers.index(key)
            else:
                raise KeyError(f"Invalid key, {key}")


    def __setitem__(self, key, val) -> None:
        from excelbird.core.function import Func
        if isinstance(key, int):
            return super().__setitem__(key, val)
        if isinstance(val, Func):
            val.kwargs['id'] = key
        else:
            val.id = key
        try:
            index = self._key_to_idx(key)
            self[index] = val
        except Exception:
            self.append(val)

    def __getitem__(self, key) -> Any:
        if not isinstance(key, slice):
            return super().__getitem__(self._key_to_idx(key))

        start, stop = key.start, key.stop
        if start is not None:
            start = self._key_to_idx(start)
        if stop is not None:
            stop = self._key_to_idx(stop)

        elems = super().__getitem__(slice(start, stop, key.step))

        if not isinstance(elems, list):
            return elems

        new = type(self)(*elems)
        for key, val in self.__dict__.items():
            if key == "_header":
                key = "header"
            if key == "_id":
                key = "id"
            setattr(new, key, val)
        return new

    def __repr__(self):
        # This shouldnt be here but I'm lazy
        return f"{type(self).__name__}({super().__repr__()})"

    def _init(self, args: list, **kwargs):
        """
        A container should finish its initialization by
        calling this function
        """
        from excelbird.core.expression import Expr
        from excelbird.core.function import Func

        list.__init__(self, list(args))

        for key, val in kwargs.items():
            setattr(self, key, val)

        Expr._set_use_ref_for_container_recursive(self)
        Expr._resolve_container_recursive(self)
        Func._resolve_container_recursive(self)

    def _insert_separator(self, separator: Gap | int | bool | dict) -> None:
        if type(separator) in [int, bool, dict]:
            if separator is True:
                separator = 1
            if isinstance(separator, int):
                separator = Gap(separator)
            elif isinstance(separator, dict):
                separator = Gap(1, **separator)

        for i, _ in reversed(list(enumerate(self))):
            if i > 0:
                self.insert(i, separator)



"""
Detailed documentation and code examples coming soon. For now, please use the class
reference page below:
"""
from __future__ import annotations
from copy import deepcopy

class Gap(int):
    """
    A spacer inside a container. The parent container will decide what to do with the Gap,
    and convert it into the appropriate type.

    Parameters
    ----------
    value : int, default 1
        The gap distance, measured in cells
    **kwargs :
        Additional keyword arguments are passed to the resulting type when the parent container creates it.    
        This lets you apply styling to the gap

    """

    def __new__(cls, value=None, *args, **kwargs):
        if value is None:
            value = 1
        # Need this because Gap's __init__ takes extra args, which int's
        # __new__ doesn't accept. So just call int's __new__ with the `value`
        # arg and ignore the extras which will be handled by Gap __init__
        return super(Gap, cls).__new__(cls, value)

    def __init__(self, value: int | None = None, fill: bool = False, is_margin: bool = False, **kwargs):
        if value is None:
            value = 1
        if len(kwargs) > 0:
            fill = True
        self.fill = fill
        self.is_margin = is_margin
        self.kwargs = kwargs
        int.__init__(value)

    def __len__(self):
        return self

    @property
    def fill_val(self):
        if self.fill is True:
            return ""
        return None

    def ref(self, inherit_style: bool = False, **kwargs) -> Gap:
        """
        Get a copy.

        `Gap` has this method because parent containers
        will call ``.ref()`` on each of their children, passing the
        same set of arguments. It's unlikely you'll ever want to call
        this method on a `Gap` directly.

        Parameters
        ----------
        inherit_style : bool, default False
            Copy the caller's style to the returned object.
        **kwargs : Any
            Extra keyword arguments are set as attributes on the returned
            object.

        Returns
        -------
        :class:`Gap <excelbird.Gap>`

        """
        if inherit_style is False:
            new = Gap(deepcopy(int(self)))
        else:
            new = deepcopy(self)

        for key, val in kwargs.items():
            new.kwargs[key] = val

        return new

    @property
    def width(self) -> int:
        return int(self)

    @property
    def height(self) -> int:
        return int(self)

    @classmethod
    def _explode_all_to_values(cls, container: list, val_type: type) -> None:
        """
        Given a container, explode each Gap to `val_type`
        with the gap's fill_val

        Mutates inplace: `container`
        """
        for i, elem in enumerate(container):
            if isinstance(elem, cls):
                gap = container.pop(i)
                for _ in range(gap):
                    container.insert(i, val_type(gap.fill_val, **gap.kwargs))

    @classmethod
    def _explode_all_to_series(
        cls, container: list, series_type: type, series_length: int
    ) -> None:
        """
        Given a container, explode each Gap to seriess of series_type filled with
        `val_type` with the Gap's fill_val

        Mutates inplace: `container`
        """
        val_type = series_type.elem_type
        for i, elem in enumerate(container):
            if isinstance(elem, cls):
                gap = container.pop(i)
                for _ in range(gap):
                    container.insert(
                        i,
                        series_type(
                            *[val_type(gap.fill_val) for _ in range(series_length)],
                            **gap.kwargs,
                        ),
                    )

    @classmethod
    def _convert_all_to_frames(
        cls, container: list, frame_type: type, series_length: int
    ) -> None:
        """
        Given a container, replace each Gap with a frame of series of cells.
        `series_length` sets the length of each resulting series.

        Mutates inplace: `container`
        """
        series_type = frame_type.elem_type
        val_type = series_type.elem_type
        for i, elem in enumerate(container):
            if isinstance(elem, cls):
                container[i] = frame_type(
                    *[
                        series_type(*[val_type(elem.fill_val) for _ in range(series_length)])
                        for _ in range(elem)
                    ],
                    **elem.kwargs,
                )

    def __repr__(self):
        return f"{type(self).__name__}({int(self)})"



"""
    pygments.styles.pastie
    ~~~~~~~~~~~~~~~~~~~~~~

    Style similar to the `pastie`_ default style.

    .. _pastie: http://pastie.caboo.se/

    :copyright: Copyright 2006-2022 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String, Error, \
     Number, Operator, Generic, Whitespace, Other, Token, Literal, Punctuation

PURPLE = "#8250df"
RED = "#CF222E"
BROWN = "#953800"  # Class names, constants
BLUE = "#0550AE"  # arithmetic operators, numbers, builtin in/and, True/F, None
DARK_BLUE = "#0A3069"  # strings, quotes, docstrings
GRAY = "#6E7781"  # comments
GREEN = "#2DA443"
ORANGE = "#DA5B0B"
BLACK = "#24292F"  # paren, colon, dot, underscore, brackets
PINK = "#E83E8C"
AQUA = "#459DB9"

class LightStyle(Style):
    """
    Style similar to the pastie default style.
    """

    styles = {
        Token:                  BLACK,
        Whitespace:             '#bbbbbb',
        Comment:                f'italic {GRAY}',
        Number:                 BLUE,
        Other:                  ORANGE,
        Keyword:                RED,
        Keyword.Constant:       BLUE,
        Keyword.Declaration:    RED,
        Keyword.Namespace:      RED,
        Keyword.Pseudo:         RED,
        Keyword.Reserved:       RED,
        Keyword.Type:           BROWN,

        Operator:               BLUE,
        Operator.Word:          BLUE,

        Name:                   BLACK,
        Name.Attribute:         RED,
        Name.Builtin:           PURPLE,
        Name.Builtin.Pseudo:    RED,
        Name.Class:             BROWN,
        Name.Constant:          BROWN,
        Name.Decorator:         PURPLE,
        Name.Entity:            BROWN,
        Name.Class:             BROWN,
        # Name.Exception:         'ne',
        Name.Function:          PURPLE,
        Name.Function.Magic:    PURPLE,
        Name.Property:          ORANGE,
        Name.Label:             ORANGE,
        Name.Namespace:         BLACK,
        Name.Other:             GREEN,
        Name.Tag:               GREEN,
        Name.Variable:          GREEN,
        Name.Variable.Class:    GREEN,
        Name.Variable.Global:   GREEN,
        Name.Variable.Instance: GREEN,
        Name.Variable.Magic:    GREEN,

        String:                 DARK_BLUE,
        String.Affix:           BLUE,
        # String.Backtick:        GREEN,
        # String.Char:            GREEN,
        # String.Delimiter:       GREEN,
        # String.Doc:             GREEN,
        # String.Double:          GREEN,
        String.Escape:          BLUE,
        # String.Heredoc:         GREEN,
        String.Interpol:        BLUE,
        # String.Other:           GREEN,
        # String.Regex:           BLUE,
        # String.Single:          GREEN,
        # String.Symbol:          GREEN,

        # Punctuation:            GREEN,
        # Punctuation.Marker:     GREEN,

        Literal:                ORANGE,

        Generic:               '#2c2cff',
        Generic.Emph:          '#008800',
        Generic.Error:         '#d30202',
        Error:                 'bg:#e3d2d2 #a61717'
    }

    # styles = {
    #     Whitespace:             '#bbbbbb',
    #     Comment:                '#888888',
    #     Comment.Preproc:        'bold #cc0000',
    #     Comment.Special:        'bg:#fff0f0 bold #cc0000',
    #
    #     String:                 'bg:#fff0f0 #dd2200',
    #     String.Regex:           'bg:#fff0ff #008800',
    #     String.Other:           'bg:#f0fff0 #22bb22',
    #     String.Symbol:          '#aa6600',
    #     String.Interpol:        '#3333bb',
    #     String.Escape:          '#0044dd',
    #
    #     Operator.Word:          '#008800',
    #
    #     Keyword:                RED,
    #     Keyword.Pseudo:         'nobold',
    #     Keyword.Type:           '#888888',
    #
    #     Name.Class:             'bold #bb0066',
    #     Name.Exception:         'bold #bb0066',
    #     Name.Function:          'bold #0066bb',
    #     Name.Property:          'bold #336699',
    #     Name.Namespace:         'bold #bb0066',
    #     Name.Builtin:           PURPLE,
    #     Name.Variable:          '#336699',
    #     Name.Variable.Class:    '#336699',
    #     Name.Variable.Instance: '#3333bb',
    #     Name.Variable.Global:   '#dd7700',
    #     Name.Constant:          'bold #003366',
    #     Name.Tag:               'bold #bb0066',
    #     Name.Attribute:         '#336699',
    #     Name.Decorator:         '#555555',
    #     Name.Label:             'italic #336699',
    #
    #     Number:                 BLUE,
    #
    #     Generic.Heading:        '#333',
    #     Generic.Subheading:     '#666',
    #     Generic.Deleted:        'bg:#ffdddd #000000',
    #     Generic.Inserted:       'bg:#ddffdd #000000',
    #     Generic.Error:          '#aa0000',
    #     Generic.Emph:           'italic',
    #     Generic.Strong:         'bold',
    #     Generic.Prompt:         '#555555',
    #     Generic.Output:         '#888888',
    #     Generic.Traceback:      '#aa0000',
    #
    #     Error:                  'bg:#e3d2d2 #a61717'
    # }




# purple - function, decorator, kwarg, module























from excelbird._base.dotdict import Style
from excelbird.styles.colors import conditional

"""
Styles
------
default_table_style:
    Used when setting ``table_style=True`` for a ``Frame``
conditional:
    Formats cells as 'good', 'bad', or 'neutral', by applying fill color and font
    color as seen in Excel's default conditional formatting styles.
"""

default_table_style = Style(
    name="TableStyleMedium2",
    showRowStripes=True,
)

conditional=Style(
    good=Style(
        fill_color=conditional.light_green,
        color=conditional.dark_green,
    ),
    bad=Style(
        fill_color=conditional.light_red,
        color=conditional.dark_red,
    ),
    neutral=Style(
        fill_color=conditional.light_yellow,
        color=conditional.dark_yellow,
    ),
),




from typing import overload, TypeVar
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.utils.cell import coordinate_from_string, column_index_from_string
from openpyxl.utils import get_column_letter

TLoc = TypeVar("TLoc", bound="Loc")

class Loc:
    """
    Stores y-x coordinates (ZERO-BASED) of a cell in Excel.
    Takes arguments in the order, `y`, `x`.
    ---
    Can be constructed from:
        String name of a cell in Excel: "B5"
        Another Loc
        2-element iterable of ints (y, x)
        Keywords, 'y' and 'x'
        2 positionals (y and x)

    Conversion to string (`str(my_loc)`) will return an Excel cell location "C7"
    """

    @overload
    def __init__(self, loc: TLoc, ws: None = None) -> None:
        ...

    @overload
    def __init__(self, loc: str | tuple[int, int] | list[int], ws: Worksheet) -> None:
        ...

    def __init__(
        self, loc: TLoc | str | tuple[int, int] | list[int], ws: Worksheet | None = None
    ) -> None:
        if isinstance(loc, Loc):
            self.y = loc.y
            self.x = loc.x
            self.ws = loc.ws

        elif isinstance(loc, (tuple, list)):
            self.y = loc[0]
            self.x = loc[1]
            self.ws = ws

        elif isinstance(loc, str):
            col_str, row_num = coordinate_from_string(loc)
            col_num = column_index_from_string(col_str)
            self.y = row_num - 1
            self.x = col_num - 1
            self.ws = ws
        else:
            raise ValueError(f"Invalid argument, {loc}")

        if self.ws is None:
            raise ValueError("A Loc must have a worksheet")

    @property
    def cell(self) -> str:
        return self.ws[self.cell_str]

    @property
    def col_letter(self) -> str:
        return get_column_letter(self.x + 1)

    @property
    def cell_str(self) -> str:
        return f"{get_column_letter(self.x+1)}{self.y+1}"

    @property
    def title_str(self) -> str:
        chars_to_trigger_quotes = [" ", "-"]
        if any(c in self.ws.title for c in chars_to_trigger_quotes):
            return "'" + self.ws.title + "'" + "!"
        return self.ws.title + "!"

    @property
    def full_str(self) -> str:
        return self.title_str + self.cell_str

    @property
    def column_dimensions(self):
        return self.ws.column_dimensions[self.col_letter]

    @property
    def row_dimensions(self):
        return self.ws.row_dimensions[self.y + 1]



"""
``__rshift__(a, b)``
    * ``a >> b`` --> ``=a:b``

``__eq__(a, b)``
    * ``a == b`` --> ``=a = b``

``__ne__(a, b)``
    * ``a != b`` --> ``=a <> b``

``__lt__(a, b)`` / ``__gt__(a, b)``
    * ``a < b`` / ``a > b`` --> ``=a < b`` / ``=a > b``

``__le__(a, b)`` / ``__ge__(a, b)``
    * ``a <= b`` / ``a >= b` --> ``=a <= b`` / ``=a >= b``

``__add__(a, b)``
    * ``a + b`` --> ``=a ^ b``

``__and__(a, b)``
    * ``a & b`` --> ``=a & b``
    * note - Strings will be surrounded in double quotes

``__sub__(a, b)``
    * ``a - b`` --> ``=a - b``

``__mul__(a, b)``
    * ``a * b`` --> ``=a * b``

``__truediv__(a, b)``
    * ``a / b`` --> ``=a / b``

``__pow__(a, b)``
    * ``a ** b`` --> ``=a ^ b``

``__xor__(a, b)``
    * ``a ^ b`` --> ``=a ^ b``

``__or__(a, b)``
    * ``a | b`` --> ``=OR(a, b)``

``__mod__(a, b)``
    * ``a % b`` --> ``=MOD(a, b)``
    * ``a % a`` --> ``=a%``
    * NOTE: Passing the *same* object on both sides of the expression (i.e. ``my_var % my_var``)
      will call excel's shorthand percent conversion and return ``=my_var%``

``__round__(a, b)``
    * ``round(a, b)`` --> ``=ROUND(a, b)``

``__abs__(x)``
    * ``abs(x)`` --> ``=ABS(x)``

``__invert__(x)``
    * ``~ x`` -> ``=NOT(x)``

``__trunc__(x)``
    * ``math.trunc(x)`` -> ``=TRUNC(x)``

``__floor__(x)``
    * ``math.floor(x)`` -> ``=FLOOR(x, 1)``

``__ceil__(x)``
    * ``math.ceil(x)`` -> ``=CEILING(x, 1)``





"""
from typing import Any
from pandas import Series, DataFrame

from excelbird._utils.util import get_dimensions
from excelbird.core.gap import Gap
from excelbird._layout_references import Globals

expr_error = ValueError(
    "You've tried to do math with an Expr that contains unresolved references. This isn't possible "
    "because to evaluate this python expression, we need to establish cell references between "
    "the two elements. Since the Expr can't be evaluated yet, we have no cells to reference! "
    "\nHowever, you can always reference another Expr inside of an Expr (that's their purpose). "
    "To do this, assign this Expr an `id` or `header` (which will be applied to its returned element) "
    "and then create a new Expr which references that identifier."
)
func_error = ValueError(
    "You've tried to do math with a Func that contains unresolved references. This isn't possible "
    "because to evaluate this python expression we need to establish cell references between the "
    "two elements. Since the Func can't be evaluated yet, we have no cells to reference! "
    "\nFortunately, you can easily fix this by using an Expr, which allows evaluation to be delayed "
    "until all of its referenecs are resolved. Simply assign this func an `id` or `header` "
    "(which will be applied to its returned element) and then reference that future element in your Expr."
)


def elem_math(a: Any, b: Any, func, sign: str) -> Any:
    from excelbird.core.cell import Cell
    from excelbird.core.series import Col
    from excelbird.core.frame import Frame
    from excelbird.core.function import Func
    from excelbird.core.expression import Expr

    a_cls, b_cls = type(a), type(b)
    a_dim, b_dim = get_dimensions(a), get_dimensions(b)

    if a_cls in [int, str, float]:
        a_dim = 0
    if b_cls in [int, str, float]:
        b_dim = 0

    if type(a) in [list, tuple, Series]:
        a_cls = Col
        a_dim = 1
    elif type(a) in [DataFrame]:
        a_cls = Frame
        a_dim = 2

    if type(b) in [list, tuple, Series]:
        b_cls = Col
        b_dim = 1
    elif type(b) in [DataFrame]:
        b_cls = Frame
        b_dim = 2
    
    if isinstance(a, Expr):
        try:
            a = a._eval()
            a_cls = type(a)
            a_dim = get_dimensions(a)
        except Exception:
            raise expr_error

    elif isinstance(a, Func):
        try:
            a = a._get_function()
            a_cls = type(a)
            a_dim = get_dimensions(a)
        except Exception:
            raise func_error

    if isinstance(b, Expr):
        try:
            b = b._eval()
            b_cls = type(b)
            b_dim = get_dimensions(b)
        except Exception:
            raise expr_error

    elif isinstance(b, Func):
        try:
            b = b._get_function()
            b_cls = type(b)
            b_dim = get_dimensions(b)
        except Exception:
            raise func_error

    assert not isinstance(a, Gap) and not isinstance(b, Gap), (
        "Cannot do math against a Gap"
    )

    if a_dim == 0 and b_dim == 0:
        return Cell(expr=[a, sign, b])

    if a_dim == b_dim:
        a = [a for a in a if not isinstance(a, Gap)]
        b = [b for b in b if not isinstance(b, Gap)]
        return a_cls(*[func(e1, e2) for e1,e2 in zip(a, b)])
    
    if a_dim > b_dim:
        a = [a for a in a if not isinstance(a, Gap)]
        return a_cls(*[func(elem, b) for elem in a])

    if a_dim < b_dim:
        b = [b for b in b if not isinstance(b, Gap)]
        if hasattr(b_cls, "sibling_type") and a_dim > 0:
            if getattr(b_cls.sibling_type, "elem_type", None) == a_cls:
                b_cls = b_cls.sibling_type

        return b_cls(*[func(a, elem) for elem in b])


class CanDoMath:

    def _space_sign(self, sign: str) -> str:
        space = Globals.expression_sign_spacing
        return (space * " ") + sign + (space * " ")

    def _to_func(self, name: str, extra_param: Any = None):
        from excelbird.core.function import Func
        start = name.upper() + "("
        if extra_param is None:
            return Func(start, self, ")")
        return Func(start, self, ", ", extra_param, ")")

    def __neg__(self):
        from excelbird.core.cell import Cell
        from excelbird.core.function import Func
        from excelbird.core.expression import Expr
        res = self
        cls = type(res)
        dim = get_dimensions(res)

        if isinstance(res, Expr):
            try:
                res = res._eval()
                dim = get_dimensions(res)
            except Exception:
                raise expr_error

        if isinstance(res, Func):
            try:
                res = res._get_function()
                dim = get_dimensions(res)
            except Exception:
                raise func_error

        if dim == 0:
            return Cell(expr=["-", res])
        
        if dim > 0:
            return cls(*[-elem for elem in res])

        assert False, (
            "Internal developer error. Can't find dimensions of CanDoMath object"
        )

    def __mod__(self, other):
        if self is not other:
            return self._to_func("MOD", other)

        from excelbird.core.cell import Cell
        from excelbird.core.function import Func
        from excelbird.core.expression import Expr
        res = self
        cls = type(res)
        dim = get_dimensions(res)

        if isinstance(res, Expr):
            try:
                res = res._eval()
                dim = get_dimensions(res)
            except Exception:
                raise expr_error

        if isinstance(res, Func):
            try:
                res = res._get_function()
                dim = get_dimensions(res)
            except Exception:
                raise func_error

        if dim == 0:
            return Cell(expr=[res, "%"])
        
        if dim > 0:
            return cls(*[elem % elem for elem in res])

        assert False, (
            "Internal developer error. Can't find dimensions of CanDoMath object"
        )

    def __round__(self, amount):
        return self._to_func("ROUND", amount)

    def __abs__(self):
        return self._to_func("ABS")

    def __invert__(self):
        return self._to_func("NOT")

    def __floor__(self):
        return self._to_func("FLOOR", 1)

    def __ceil__(self):
        return self._to_func("CEILING", 1)

    def __trunc__(self):
        return self._to_func("TRUNC")

    def __or__(self, other):
        return self._to_func("OR", other)

    def __ror__(self, other):
        from excelbird.core.function import Func
        return Func("OR(", other, ", ", self, ")")

    def __and__(self, other):
        return self._to_func("AND", other)

    def __rand__(self, other):
        from excelbird.core.function import Func
        return Func("AND(", other, ", ", self, ")")

    def __eq__(self, other):
        return elem_math(self, other, lambda a,b: a == b, self._space_sign("="))

    def __ne__(self, other):
        return elem_math(self, other, lambda a,b: a != b, self._space_sign("<>"))

    def __lt__(self, other):
        return elem_math(self, other, lambda a,b: a < b, self._space_sign("<"))

    def __gt__(self, other):
        return elem_math(self, other, lambda a,b: a > b, self._space_sign(">"))

    def __le__(self, other):
        return elem_math(self, other, lambda a,b: a <= b, self._space_sign("<="))

    def __ge__(self, other):
        return elem_math(self, other, lambda a,b: a >= b, self._space_sign(">="))

    def __add__(self, other):
        from excelbird.core.function import Func
        if isinstance(self, Func) and self._is_python_sum is True:
            self.inner.insert(-1, ", ")
            self.inner.insert(-1, other)
            return self

        if isinstance(other, str):
            if not other.startswith('"') and not other.endswith('"'):
                other = f'"{other}"'
            return elem_math(self, other, lambda a,b: a + b, self._space_sign("&"))
        return elem_math(self, other, lambda a,b: a + b, self._space_sign("+"))

    def __radd__(self, other):
        from excelbird.core.function import Func
        if other == 0:  # Adding to zero signals the beginning of python sum() evaluation
            return Func("SUM(", self, ")", _is_python_sum = True)

        if isinstance(other, Func) and other._is_python_sum is True:
            other.inner.insert(-1, ", ")
            other.inner.insert(-1, self)
            return other

        if isinstance(other, str):
            if not other.startswith('"') and not other.endswith('"'):
                other = f'"{other}"'
            return elem_math(other, self, lambda a,b: a + b, self._space_sign("&"))
        return elem_math(other, self, lambda a,b: a + b, self._space_sign("+"))

    def __sub__(self, other):
        return elem_math(self, other, lambda a,b: a - b, self._space_sign("-"))

    def __rsub__(self, other):
        return elem_math(other, self, lambda a,b: a - b, self._space_sign("-"))


    def __mul__(self, other):
        return elem_math(self, other, lambda a,b: a * b, self._space_sign("*"))

    def __rmul__(self, other):
        return elem_math(other, self, lambda a,b: a * b, self._space_sign("*"))


    def __truediv__(self, other):
        return elem_math(self, other, lambda a,b: a / b, self._space_sign("/"))

    def __rtruediv__(self, other):
        return elem_math(other, self, lambda a,b: a / b, self._space_sign("/"))


    def __xor__(self, other):
        return elem_math(self, other, lambda a,b: a ^ b, self._space_sign("^"))

    def __rxor__(self, other):
        return elem_math(other, self, lambda a,b: a ^ b, self._space_sign("^"))


    def __pow__(self, other):
        return elem_math(self, other, lambda a,b: a ^ b, self._space_sign("^"))

    def __rpow__(self, other):
        return elem_math(other, self, lambda a,b: a ^ b, self._space_sign("^"))


    def __rshift__(self, other):
        return elem_math(self, other, lambda a,b: a >> b, ":")

    def __rrshift__(self, other):
        return elem_math(self, other, lambda a,b: a >> b, ":")


    # def __and__(self, other):
    #     if isinstance(other, str):
    #         if not other.endswith('"') and not other.startswith('"'):
    #             other = f'"{other}"'
    #     return elem_math(self, other, lambda a,b: a & b, self._space_sign("&"))
    #
    # def __rand__(self, other):
    #     if isinstance(other, str):
    #         if not other.endswith('"') and not other.startswith('"'):
    #             other = f'"{other}"'
    #     return elem_math(self, other, lambda a,b: a & b, self._space_sign("&"))



"""
A :class:`Func <excelbird.Func>` is a template that dynamically returns an
element of the appropriate shape/dimensions while applying an Excel formula to its
cell(s). It should be used **only when calling builtin Excel functions** in your formula.

As standard across excelbird, a Func is designed for use without
specifying any cell coordinates.

Syntax
------
To build a :class:`Func <excelbird.Func>`, provide the *exact* formula string,
with Python objects placed wherever the cell references should appear.

For instance, here's how it should look if you were **not** referencing any cells.

.. code-block::

    from excelbird import Func, Expr, Cell, Col, Frame, Stack

    simple_func = Func("SUM(5, 4, MIN(6, 7))")

To reference other layout elements in your formula, there are two options:

Referencing Elements - Option 1
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
**Break the string, and insert variables directly**. Be warned, this option
can be clumsy, tedious, and cause mistakes.

Let's rewrite the above example, but instead of
hardcoding `5` and `4`, reference :class:`Cells <excelbird.Cell>` containing them.

.. code-block::

    five = Cell(5)
    four = Cell(4)
    another_func = Func("SUM(", five, ", ", four, ", MIN(6, 7))")

Notice the comma we had to insert manually between `five` and `four`

.. note::

    The above technique can be clumsy and prone to user-error. Its only
    benefit is improved speed, and that it doesn't require any knowledge
    of object `ids`, `headers`, and :class:`Expr <excelbird.Expr>`.

Referencing Elements - Option 2
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
**Templated** :class:`Expr <excelbird.Expr>`. Please read the intro to `Expr` if
you aren't familiar with them yet.

.. code-block::

    better_func = Func(f"SUM({{cell_five}}, {{cell_four}}, MIN(6, 7))")
    five = Cell(5, id="cell_five")  # ^ Exprs can reference stuff that doesn't exist yet.
    four = Cell(4, id="cell_four")

`Why` did we use braces ``{{...}}`` instead of square bracktets ``[...]`` as we do inside an `Expr`?

**In an Expr, square brackets enclose an id or header. In a Func, curly braces enclose an Expr**

To illustrate this, let's make our Expr execute some python code. We'll multiply cells five and four
together, instead of listing them. For reference, in **option 1** this would look like
``Func("SUM(", five * four, ", MIN(6, 7))")``

.. code-block::

    Func(f"SUM({{[cell_five] * [cell_four]}}, MIN(6, 7))")

Which is equivalent to

.. code-block::

    Func("SUM(", Expr("[cell_five] * [cell_four]"), ", MIN(6, 7))")

**Why are the braces necessary? Why can't I just use square brackets alone?**

Because in an Expr, the entire string is treated as *valid Python code* and **is executed**.
This allows for ``Expr("[row_header].range() + [col_header].loc[3:]")``. If this
were placed inside a Func, it couldn't be determined whether the ``.range``, ``.loc``,
or ``+`` should be executed in Python, or placed as exact strings in the Excel formula. Excelbird
could guess, but if it guessed incorrectly, you'd have a corrupted file.

.. note::

    **Double braces are strongly recommended**, but **any** number of curly braces are allowed. i.e. ``{{{stuff}}}``, ``{{stuff}}``,
    or ``{stuff}`` are each valid. This is a design feature to ensure your code will look the same regardless of whether
    the string is `an f-string <https://www.geeksforgeeks.org/formatted-string-literals-f-strings-python/>`_.


Additional Arguments - res_type
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

:attr:`res_type` can be used to specify the type of the returned element.
This is **not** necessary inside structured parent containers like :class:`Frame <excelbird.Frame>`
or :class:`Row <excelbird.Row>` which decide automatically. Inside `unstructured` containers
like :class:`Stack <excelbird.Stack>` or :class:`Sheet <excelbird.Sheet>` it must be specified manually

For instance, in the following example, the parent Frame knows that ``res_type`` **must** be
:class:`Col <excelbird.Col>`

.. code-block::

    Frame(
        Func(f"MAX({{second}}, {{third}})", header="first"),
        Col(1, 2, 3, header="second"),
        Col(2, 7, 1, header="third"),
    )

If those elements were placed in a Stack, we have to decide:
Should it return a Col with the max of each pair in the other vectors?
Or should it return a Cell with the single largest value in either vector? Try both:

.. code-block::

    Stack(
        Col(1, 2, 3, header="first"),
        Col(2, 7, 1, header="second"),
        Func(f"MAX({{first}}, {{second}})", header="third", res_type=Col),
        Func(f"MAX({{first}}, {{second}})", res_type=Cell),
    )


----

Here is an **impractical** example that demonstrates a variety of syntax at the same time.

.. code-block::

    my_cell = Cell(69)
    Func("SUM(", {"foo"}, ", ", my_cell, None, ", ", Expr("bar"), f", SUM({{[one] * [two]}}, {{some_elem}}) + ", 5, ")")

will build a formula that's arranged like:

.. code-block::

    =SUM(<Expr>, 99, <Cell>, SUM(<Expr>, <Expr>) + 5)

where the second `Expr` is the result of ``Expr("[one] * [two]")`` and the
third `Expr` is the result of ``Expr("some_elem")``. Notice the value of
``None`` was ignored, which is a feature shared by all excelbird layout elements
to allow for conditional placement of elements.

"""
from __future__ import annotations
import re
from typing import Any
from itertools import zip_longest
from itertools import chain

from excelbird.core.expression import Expr
from excelbird._base.math import CanDoMath

from excelbird._utils.util import (
    get_dimensions,
)
from excelbird._utils.argument_parsing import (
    convert_all_to_type
)

class Func(CanDoMath):
    """
    Create a formula that uses builtin Excel functions. One `Func` does *not* correspond
    to a single call to an Excel function, but rather to the entire formula. Therefore, you
    cannot nest ``Func`` inside one another, nor would there be any reason to.

    .. note::

        `Func` should *only* be used if calling **builtin Excel functions**. If just
        arithmetic is needed to build your formula (i.e. adding/subtracting things), you
        can execute the expression **directly in Python** on layout elements, and optionally
        do so by referencing elements by id/header using an :class:`Expr`.

    Parameters
    ----------
    *inner : Any
        Pass each section of the formula as positional arguments
    res_type : type, optional
        Desired return type. This is rarely needed, since often you'll be placing ``Func``
        inside a container that can only hold one type, so ``res_type`` is inferred.
        In some cases though, such as when inside a ``Stack``, or when being used in a
        math expression prior to being placed in a layout, this will need to be specified.
    **kwargs : Any
        Any additional keyword arguments will be passed to the constructor of the resulting
        object.

    """

    def __init__(
        self,
        *inner,
        res_type: type | None = None,
        _is_python_sum: bool = False,
        **kwargs,
    ) -> None:
        """
        All that is needed is the res type and a list of elements.
        Define the function in its exact form, including all characters,
        including parentheses, except the beginnning equals sign. Separate
        out the dynamic elements into their own list elements, which can either
        be an Expression, set, Cell or series.

        If res type is a cell, call .range()
        """
        if len(inner) == 1:
            if type(inner[0]) in (list, tuple):
                inner = inner[0]

        inner = list(inner)

        inner = self._parse_args(inner)

        if len(inner) == 0:
            raise ValueError("No elements provided to Func")

        if any(isinstance(i, Func) for i in inner):
            raise ValueError("Cannot nest Funcs inside one another, as there is no need to do so.")

        self.res_type = res_type
        self.inner = inner
        self._is_python_sum = _is_python_sum
        self.kwargs = kwargs

    def set(self, **kwargs) -> Func:
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                self.kwargs[k] = v
        return self

    def _parse_args(self, args) -> list:
        def extract_exprs_from_string(s: str):
            matches = [Expr(s) for s in re.findall(r"\{+(.*?)\}+", s)]
            splits = re.split(r"\{+.*?\}+", s)
            sub_args = [x for x in chain.from_iterable(zip_longest(splits, matches)) if x]
            return sub_args

        def format_string(s: str) -> str:
            # Remove comments from end of string
            s = re.sub(r"#+.*?$", "", s)
            # Remove comments at the end of lines
            s = re.sub(r"#+.*?\n", " ", s)
            # Remove whitespace
            s = re.sub(r"\s+", " ", s)
            return s

        def is_str_containing_expr(s: Any) -> bool:
            if isinstance(s, str):
                if "{" in s:
                    return True
            return False

        args = [format_string(i) if isinstance(i, str) else i for i in args]

        return list(
            chain.from_iterable(
                [
                    [Expr(elem.pop())] if isinstance(elem, set)
                    else [elem] if not is_str_containing_expr(elem)
                    else extract_exprs_from_string(elem)
                    for elem in args if elem is not None
                ]
            )
        )


    def _get_function(self, container_type: type | None = None):
        if self.res_type is None:
            if container_type is None:
                raise ValueError(
                    "Can't determine the result type of function. Please provide "
                    "`res_type` as a keyword argument to Func, with the desired return type."
                )
            else:
                from excelbird.core.frame import Frame, VFrame
                from excelbird.core.series import Col, Row
                valid_containers = (Frame, VFrame, Col, Row)
                if not issubclass(container_type, valid_containers):
                    raise ValueError(
                        f"When a Func is placed inside of a general container type "
                        f"(in this case, {container_type.__name__}) we can't assume  "
                        "its result type cannot be implied/assumed (we don't know) if "
                    )
                self.res_type = container_type.elem_type

        assert self.res_type is not None, (
            "Internal developer error: a Func's res_type is still None. Why? Please raise an Issue on Github so this can be patched. "
            "In the meantime, you can avoid this error by passing `res_type` as a keyword argument to Func"
        )

        if self._all_resolved() is False:
            raise ValueError("All references must be resolved before calling .get_function()")

        dimensions = get_dimensions(self.res_type)
        assert dimensions >= 0, f"Invalid res_type, {self.res_type}"

        if dimensions == 0:
            # Since we're only returning one cell, we need to take each element
            # which is larger than a cell, and only refer to its range
            for i, elem in enumerate(self.inner):
                if get_dimensions(elem) > 0:
                    self.inner[i] = elem.range()
            return self.res_type(func=self.inner, **self.kwargs)

        if dimensions == 1:
            res_length = min(len(x) for x in self.inner if get_dimensions(x) == 1)
            elem_type = self.res_type.elem_type
            for i, elem in enumerate(self.inner):
                if get_dimensions(elem) > 1:
                    frame = elem
                    self.inner[i] = self.res_type(
                        *[ frame[0][k] >> frame[-1][k] for k in range(res_length) ]
                    )
            return self.res_type(
                *[
                    elem_type(func=[item[i] if get_dimensions(item) == 1 else item for item in self.inner])
                    for i in range(res_length)
                ],
                **self.kwargs
            )

        if dimensions == 2:
            raise ValueError(
                "Returning a Frame from Func is not allowed yet. Please return a Cell, Row or Col."
            )

        raise Exception("get function returned nothing")


    def _exprs(self) -> list:
        return [i for i in self.inner if isinstance(i, Expr)]

    def _all_resolved(self) -> bool:
        if any(i._refs_resolved() is False for i in self._exprs()):
            return False
        return True
    
    def _attempt_to_resolve(self, container: list) -> bool:
        """
        Given a parent container, try to resolve each expression
        in self's `inner`.
        """
        for i, elem in enumerate(self.inner):
            if isinstance(elem, Expr):
                if elem._attempt_to_resolve(container) is True:
                    self.inner[i] = elem._eval()

        return self._all_resolved()

    @classmethod
    def _resolve_container_recursive(cls, container: list) -> bool:
        """
        For each _DelayedFunc in a container, attempt to resolve each of
        its expressions.

        Returns True if all were resolved

        Mutates inplace: `container`
        """
        all_dfuncs_resolved = True

        for i, elem in enumerate(container):
            if isinstance(elem, cls):

                if elem._attempt_to_resolve(container) is True:
                    container[i] = elem._get_function(type(container))
                else:
                    all_dfuncs_resolved = False

            elif isinstance(elem, list):
                if cls._resolve_container_recursive(elem) is False:
                    all_dfuncs_resolved = False

        return all_dfuncs_resolved

    def __repr__(self):
        if self.res_type is not None:
            return f"{type(self).__name__}({self.res_type.__name__}...)"
        return f"{type(self).__name__}(...)"




from excelbird.core import *
from excelbird.styles import *
from excelbird.schema import *
from excelbird.exceptions import *
from excelbird._layout_references import *
from excelbird._base.dotdict import *



from excelbird._base.container import *
from excelbird._base.dotdict import *
from excelbird._base.help import *
from excelbird._base.identifier import *
from excelbird._base.loc import *
from excelbird._base.math import *
from excelbird._base.styling import *
